@@ -12,17 +12,21 @@ use std::io::Result;
 use std::io::SeekFrom;
 use std::os::unix::fs::{chown, fchown, lchown, symlink};
 use std::path::{Path, PathBuf};
-use std::process::ChildStdout;
-use std::process::Command;
-use std::process::Stdio;
 use std::time::SystemTime;
 
-use crate::header::*;
+use crate::compression::Compression;
+use crate::filetype::*;
+use crate::header::Header;
 use crate::libc::{mknod, set_modified, strftime_local};
+use crate::manifest::Manifest;
 use crate::seek_forward::SeekForward;
 
+mod compression;
+mod extended_error;
+mod filetype;
 mod header;
 mod libc;
+mod manifest;
 mod seek_forward;
 
 pub const LOG_LEVEL_WARNING: u32 = 5;
@@ -30,14 +34,14 @@ pub const LOG_LEVEL_INFO: u32 = 7;
 pub const LOG_LEVEL_DEBUG: u32 = 8;
 
 struct CpioFilenameReader<'a, R: Read + SeekForward> {
-    file: &'a mut R,
+    archive: &'a mut R,
 }
 
 impl<R: Read + SeekForward> Iterator for CpioFilenameReader<'_, R> {
     type Item = Result<String>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        match read_filename_from_next_cpio_object(self.file) {
+        match read_filename_from_next_cpio_object(self.archive) {
             Ok(filename) => {
                 if filename == "TRAILER!!!" {
                     None
@@ -121,7 +125,7 @@ impl Extractor {
     fn set_modified_times(&self, log_level: u32) -> Result<()> {
         for (path, mtime) in self.mtimes.iter().rev() {
             if log_level >= LOG_LEVEL_DEBUG {
-                writeln!(std::io::stderr(), "set mtime {} for '{}'", mtime, path)?;
+                writeln!(std::io::stderr(), "set mtime {mtime} for '{path}'")?;
             };
             set_modified(path, *mtime)?;
         }
@@ -142,14 +146,14 @@ fn align_to_4_bytes(length: u32) -> u32 {
 ///
 /// Read the next cpio object header, check the magic, skip the file data.
 /// Return the file name.
-fn read_filename_from_next_cpio_object<R: Read + SeekForward>(file: &mut R) -> Result<String> {
-    let (filesize, filename) = Header::read_only_filesize_and_filename(file)?;
+fn read_filename_from_next_cpio_object<R: Read + SeekForward>(archive: &mut R) -> Result<String> {
+    let (filesize, filename) = Header::read_only_filesize_and_filename(archive)?;
     let skip = filesize + align_to_4_bytes(filesize);
-    file.seek_forward(skip.into())?;
+    archive.seek_forward(skip.into())?;
     Ok(filename)
 }
 
-fn read_magic_header<R: Read + Seek>(file: &mut R) -> Option<Result<Command>> {
+fn read_magic_header<R: Read + Seek>(file: &mut R) -> Option<Result<Compression>> {
     let mut buffer = [0; 4];
     while buffer == [0, 0, 0, 0] {
         match file.read_exact(&mut buffer) {
@@ -160,109 +164,46 @@ fn read_magic_header<R: Read + Seek>(file: &mut R) -> Option<Result<Command>> {
             },
         };
     }
-    let command = match buffer {
-        [0x42, 0x5A, 0x68, _] => {
-            let mut cmd = Command::new("bzip2");
-            cmd.arg("-cd");
-            cmd
-        }
-        [0x30, 0x37, 0x30, 0x37] => Command::new("cpio"),
-        [0x1F, 0x8B, _, _] => {
-            let mut cmd = Command::new("gzip");
-            cmd.arg("-cd");
-            cmd
-        }
-        // Different magic numbers (little endian) for lz4:
-        // v0.1-v0.9: 0x184C2102
-        // v1.0-v1.3: 0x184C2103
-        // v1.4+: 0x184D2204
-        [0x02, 0x21, 0x4C, 0x18] | [0x03, 0x21, 0x4C, 0x18] | [0x04, 0x22, 0x4D, 0x18] => {
-            let mut cmd = Command::new("lz4");
-            cmd.arg("-cd");
-            cmd
-        }
-        [0x5D, _, _, _] => {
-            let mut cmd = Command::new("lzma");
-            cmd.arg("-cd");
-            cmd
-        }
-        // Full magic number for lzop: [0x89, 0x4C, 0x5A, 0x4F, 0x00, 0x0D, 0x0A, 0x1A, 0x0A]
-        [0x89, 0x4C, 0x5A, 0x4F] => {
-            let mut cmd = Command::new("lzop");
-            cmd.arg("-cd");
-            cmd
-        }
-        // Full magic number for xz: [0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00]
-        [0xFD, 0x37, 0x7A, 0x58] => {
-            let mut cmd = Command::new("xz");
-            cmd.arg("-cd");
-            cmd
-        }
-        [0x28, 0xB5, 0x2F, 0xFD] => {
-            let mut cmd = Command::new("zstd");
-            cmd.arg("-cdq");
-            cmd
-        }
-        _ => {
-            return Some(Err(Error::new(
-                ErrorKind::InvalidData,
-                format!(
-                    "Failed to determine CPIO or compression magic number: 0x{:02x}{:02x}{:02x}{:02x} (big endian)",
-                    buffer[0], buffer[1], buffer[2], buffer[3]
-                ),
-            )));
-        }
-    };
     match file.seek(SeekFrom::Current(-4)) {
         Ok(_) => {}
         Err(e) => {
             return Some(Err(e));
         }
     };
-    Some(Ok(command))
-}
-
-fn decompress(command: &mut Command, file: File) -> Result<ChildStdout> {
-    // TODO: Propper error message if spawn fails
-    let cmd = command
-        .stdin(file)
-        .stdout(Stdio::piped())
-        .spawn()
-        .map_err(|e| match e.kind() {
-            ErrorKind::NotFound => Error::other(format!(
-                "Program '{}' not found in PATH.",
-                command.get_program().to_str().unwrap()
-            )),
-            _ => e,
-        })?;
-    // TODO: Should unwrap be replaced by returning Result?
-    Ok(cmd.stdout.unwrap())
+    let compression = match Compression::from_magic_number(buffer) {
+        Ok(compression) => compression,
+        Err(e) => {
+            return Some(Err(e));
+        }
+    };
+    Some(Ok(compression))
 }
 
 fn read_cpio_and_print_filenames<R: Read + SeekForward, W: Write>(
-    file: &mut R,
+    archive: &mut R,
     out: &mut W,
 ) -> Result<()> {
-    let cpio = CpioFilenameReader { file };
+    let cpio = CpioFilenameReader { archive };
     for f in cpio {
         let filename = f?;
-        writeln!(out, "{}", filename)?;
+        writeln!(out, "{filename}")?;
     }
     Ok(())
 }
 
 fn read_cpio_and_print_long_format<R: Read + SeekForward, W: Write>(
-    file: &mut R,
+    archive: &mut R,
     out: &mut W,
     now: i64,
     user_group_cache: &mut UserGroupCache,
+    print_ino: bool,
 ) -> Result<()> {
     // Files can have the same mtime (especially when using SOURCE_DATE_EPOCH).
     // Cache the time string of the last mtime.
     let mut last_mtime = 0;
     let mut time_string: String = "".into();
     loop {
-        let header = match Header::read(file) {
+        let header = match Header::read(archive) {
             Ok(header) => {
                 if header.filename == "TRAILER!!!" {
                     break;
@@ -287,9 +228,12 @@ fn read_cpio_and_print_long_format<R: Read + SeekForward, W: Write>(
             time_string = format_time(header.mtime, now)?;
         };
 
+        if print_ino {
+            write!(out, "{:>4} ", header.ino)?;
+        }
         match header.mode & MODE_FILETYPE_MASK {
             FILETYPE_SYMLINK => {
-                let target = header.read_symlink_target(file)?;
+                let target = header.read_symlink_target(archive)?;
                 writeln!(
                     out,
                     "{} {:>3} {:<8} {:<8} {:>8} {} {} -> {}",
@@ -304,7 +248,7 @@ fn read_cpio_and_print_long_format<R: Read + SeekForward, W: Write>(
                 )?;
             }
             FILETYPE_BLOCK_DEVICE | FILETYPE_CHARACTER_DEVICE => {
-                header.skip_file_content(file)?;
+                header.skip_file_content(archive)?;
                 writeln!(
                     out,
                     "{} {:>3} {:<8} {:<8} {:>3}, {:>3} {} {}",
@@ -319,7 +263,7 @@ fn read_cpio_and_print_long_format<R: Read + SeekForward, W: Write>(
                 )?;
             }
             _ => {
-                header.skip_file_content(file)?;
+                header.skip_file_content(archive)?;
                 writeln!(
                     out,
                     "{} {:>3} {:<8} {:<8} {:>8} {} {}",
@@ -434,7 +378,7 @@ fn from_mtime(mtime: u32) -> SystemTime {
 }
 
 fn write_file<R: Read + SeekForward>(
-    cpio_file: &mut R,
+    archive: &mut R,
     header: &Header,
     preserve_permissions: bool,
     seen_files: &mut SeenFiles,
@@ -487,7 +431,7 @@ fn write_file<R: Read + SeekForward>(
         file = File::create(&header.filename)?
     };
     header.mark_seen(seen_files);
-    let mut reader = cpio_file.take(header.filesize.into());
+    let mut reader = archive.take(header.filesize.into());
     // TODO: check writing hard-link with length == 0
     // TODO: check overwriting existing files/hardlinks
     let written = std::io::copy(&mut reader, &mut file)?;
@@ -498,7 +442,7 @@ fn write_file<R: Read + SeekForward>(
         )));
     }
     let skip = align_to_4_bytes(header.filesize);
-    cpio_file.seek_forward(skip.into())?;
+    archive.seek_forward(skip.into())?;
     if preserve_permissions {
         fchown(&file, Some(header.uid), Some(header.gid))?;
     }
@@ -508,12 +452,12 @@ fn write_file<R: Read + SeekForward>(
 }
 
 fn write_symbolic_link<R: Read + SeekForward>(
-    cpio_file: &mut R,
+    archive: &mut R,
     header: &Header,
     preserve_permissions: bool,
     log_level: u32,
 ) -> Result<()> {
-    let target = header.read_symlink_target(cpio_file)?;
+    let target = header.read_symlink_target(archive)?;
     if log_level >= LOG_LEVEL_DEBUG {
         writeln!(
             std::io::stderr(),
@@ -565,7 +509,7 @@ where
         // TODO: Use ErrorKind::InvalidFilename once stable.
         None => Err(Error::new(
             ErrorKind::InvalidData,
-            format!("Path {:#?} has no parent directory.", abspath),
+            format!("Path {abspath:#?} has no parent directory."),
         )),
     }
 }
@@ -580,8 +524,8 @@ fn check_path_is_canonical_subdir<S: AsRef<str> + std::fmt::Display>(
         return Err(Error::new(
             ErrorKind::InvalidData,
             format!(
-                "The parent directory of \"{}\" (resolved to {:#?}) is not within the directory {:#?}.",
-                path, canonicalized_path, base_dir
+                "The parent directory of \"{path}\" (resolved to {canonicalized_path:#?}) \
+                is not within the directory {base_dir:#?}.",
             ),
         ));
     }
@@ -589,7 +533,7 @@ fn check_path_is_canonical_subdir<S: AsRef<str> + std::fmt::Display>(
 }
 
 fn read_cpio_and_extract<R: Read + SeekForward>(
-    file: &mut R,
+    archive: &mut R,
     base_dir: &PathBuf,
     preserve_permissions: bool,
     log_level: u32,
@@ -597,7 +541,7 @@ fn read_cpio_and_extract<R: Read + SeekForward>(
     let mut extractor = Extractor::new();
     let mut previous_checked_dir = PathBuf::new();
     loop {
-        let header = match Header::read(file) {
+        let header = match Header::read(archive) {
             Ok(header) => {
                 if header.filename == "TRAILER!!!" {
                     break;
@@ -609,7 +553,7 @@ fn read_cpio_and_extract<R: Read + SeekForward>(
         };
 
         if log_level >= LOG_LEVEL_DEBUG {
-            writeln!(std::io::stderr(), "{:?}", header)?;
+            writeln!(std::io::stderr(), "{header:?}")?;
         } else if log_level >= LOG_LEVEL_INFO {
             writeln!(std::io::stderr(), "{}", header.filename)?;
         }
@@ -636,14 +580,14 @@ fn read_cpio_and_extract<R: Read + SeekForward>(
                 &mut extractor.mtimes,
             )?,
             FILETYPE_REGULAR_FILE => write_file(
-                file,
+                archive,
                 &header,
                 preserve_permissions,
                 &mut extractor.seen_files,
                 log_level,
             )?,
             FILETYPE_SYMLINK => {
-                write_symbolic_link(file, &header, preserve_permissions, log_level)?
+                write_symbolic_link(archive, &header, preserve_permissions, log_level)?
             }
             FILETYPE_FIFO | FILETYPE_BLOCK_DEVICE | FILETYPE_SOCKET => {
                 unimplemented!(
@@ -666,24 +610,24 @@ fn read_cpio_and_extract<R: Read + SeekForward>(
     Ok(())
 }
 
-fn seek_to_cpio_end(file: &mut File) -> Result<()> {
-    let cpio = CpioFilenameReader { file };
+fn seek_to_cpio_end(archive: &mut File) -> Result<()> {
+    let cpio = CpioFilenameReader { archive };
     for f in cpio {
         f?;
     }
     Ok(())
 }
 
-pub fn get_cpio_archive_count(file: &mut File) -> Result<u32> {
+pub fn get_cpio_archive_count(archive: &mut File) -> Result<u32> {
     let mut count = 0;
     loop {
-        let command = match read_magic_header(file) {
+        let compression = match read_magic_header(archive) {
             None => return Ok(count),
             Some(x) => x?,
         };
         count += 1;
-        if command.get_program() == "cpio" {
-            seek_to_cpio_end(file)?;
+        if compression.is_uncompressed() {
+            seek_to_cpio_end(archive)?;
         } else {
             break;
         }
@@ -691,26 +635,60 @@ pub fn get_cpio_archive_count(file: &mut File) -> Result<u32> {
     Ok(count)
 }
 
-pub fn print_cpio_archive_count<W: Write>(mut file: File, out: &mut W) -> Result<()> {
-    let count = get_cpio_archive_count(&mut file)?;
-    writeln!(out, "{}", count)?;
+// Parse SOURCE_DATE_EPOCH environment variable (if set and valid integer)
+fn get_source_date_epoch() -> Option<u32> {
+    match std::env::var("SOURCE_DATE_EPOCH") {
+        Ok(value) => match value.parse::<i64>() {
+            Ok(source_date_epoch) => {
+                if let Ok(x) = source_date_epoch.try_into() {
+                    Some(x)
+                } else if source_date_epoch < 0 {
+                    Some(0)
+                } else {
+                    Some(u32::MAX)
+                }
+            }
+            Err(_) => None,
+        },
+        Err(_) => None,
+    }
+}
+
+pub fn print_cpio_archive_count<W: Write>(mut archive: File, out: &mut W) -> Result<()> {
+    let count = get_cpio_archive_count(&mut archive)?;
+    writeln!(out, "{count}")?;
     Ok(())
 }
 
-pub fn examine_cpio_content<W: Write>(mut file: File, out: &mut W) -> Result<()> {
+pub fn create_cpio_archive(archive: Option<File>, log_level: u32) -> Result<()> {
+    let source_date_epoch = get_source_date_epoch();
+    let stdin = std::io::stdin();
+    let buf_reader = std::io::BufReader::new(stdin);
+    if log_level >= LOG_LEVEL_DEBUG {
+        eprintln!("Parsing manifest from stdin...");
+    }
+    let manifest = Manifest::from_input(buf_reader, log_level)?;
+    if log_level >= LOG_LEVEL_DEBUG {
+        eprintln!("Writing cpio...");
+    }
+    manifest.write_archive(archive, source_date_epoch, log_level)?;
+    Ok(())
+}
+
+pub fn examine_cpio_content<W: Write>(mut archive: File, out: &mut W) -> Result<()> {
     loop {
-        let command = match read_magic_header(&mut file) {
+        let compression = match read_magic_header(&mut archive) {
             None => return Ok(()),
             Some(x) => x?,
         };
         writeln!(
             out,
             "{}\t{}",
-            file.stream_position()?,
-            command.get_program().to_str().unwrap()
+            archive.stream_position()?,
+            compression.command()
         )?;
-        if command.get_program() == "cpio" {
-            seek_to_cpio_end(&mut file)?;
+        if compression.is_uncompressed() {
+            seek_to_cpio_end(&mut archive)?;
         } else {
             break;
         }
@@ -719,7 +697,7 @@ pub fn examine_cpio_content<W: Write>(mut file: File, out: &mut W) -> Result<()>
 }
 
 pub fn extract_cpio_archive(
-    mut file: File,
+    mut archive: File,
     preserve_permissions: bool,
     subdir: Option<String>,
     log_level: u32,
@@ -727,26 +705,21 @@ pub fn extract_cpio_archive(
     let mut count = 1;
     let base_dir = std::env::current_dir()?;
     loop {
+        let mut dir = base_dir.clone();
         if let Some(ref s) = subdir {
-            let mut dir = base_dir.clone();
             dir.push(format!("{s}{count}"));
             create_dir_ignore_existing(&dir)?;
             std::env::set_current_dir(&dir)?;
         }
-        let mut command = match read_magic_header(&mut file) {
+        let compression = match read_magic_header(&mut archive) {
             None => return Ok(()),
             Some(x) => x?,
         };
-        if command.get_program() == "cpio" {
-            read_cpio_and_extract(&mut file, &base_dir, preserve_permissions, log_level)?;
+        if compression.is_uncompressed() {
+            read_cpio_and_extract(&mut archive, &dir, preserve_permissions, log_level)?;
         } else {
-            let mut decompressed = decompress(&mut command, file)?;
-            read_cpio_and_extract(
-                &mut decompressed,
-                &base_dir,
-                preserve_permissions,
-                log_level,
-            )?;
+            let mut decompressed = compression.decompress(archive)?;
+            read_cpio_and_extract(&mut decompressed, &dir, preserve_permissions, log_level)?;
             break;
         }
         count += 1;
@@ -754,7 +727,7 @@ pub fn extract_cpio_archive(
     Ok(())
 }
 
-pub fn list_cpio_content<W: Write>(mut file: File, out: &mut W, log_level: u32) -> Result<()> {
+pub fn list_cpio_content<W: Write>(mut archive: File, out: &mut W, log_level: u32) -> Result<()> {
     let mut user_group_cache = UserGroupCache::new();
     let now = SystemTime::now()
         .duration_since(SystemTime::UNIX_EPOCH)
@@ -763,24 +736,31 @@ pub fn list_cpio_content<W: Write>(mut file: File, out: &mut W, log_level: u32)
         .try_into()
         .unwrap();
     loop {
-        let mut command = match read_magic_header(&mut file) {
+        let compression = match read_magic_header(&mut archive) {
             None => return Ok(()),
             Some(x) => x?,
         };
-        if command.get_program() == "cpio" {
+        if compression.is_uncompressed() {
             if log_level >= LOG_LEVEL_INFO {
-                read_cpio_and_print_long_format(&mut file, out, now, &mut user_group_cache)?;
+                read_cpio_and_print_long_format(
+                    &mut archive,
+                    out,
+                    now,
+                    &mut user_group_cache,
+                    log_level >= LOG_LEVEL_DEBUG,
+                )?;
             } else {
-                read_cpio_and_print_filenames(&mut file, out)?;
+                read_cpio_and_print_filenames(&mut archive, out)?;
             }
         } else {
-            let mut decompressed = decompress(&mut command, file)?;
+            let mut decompressed = compression.decompress(archive)?;
             if log_level >= LOG_LEVEL_INFO {
                 read_cpio_and_print_long_format(
                     &mut decompressed,
                     out,
                     now,
                     &mut user_group_cache,
+                    log_level >= LOG_LEVEL_DEBUG,
                 )?;
             } else {
                 read_cpio_and_print_filenames(&mut decompressed, out)?;
@@ -799,6 +779,9 @@ mod tests {
     use super::*;
     use crate::libc::{major, minor};
 
+    // Lock for tests that rely on / change the current directory
+    pub static TEST_LOCK: std::sync::Mutex<u32> = std::sync::Mutex::new(0);
+
     fn getgid() -> u32 {
         unsafe { ::libc::getgid() }
     }
@@ -826,7 +809,7 @@ mod tests {
             let name = std::option_env!("CARGO_PKG_NAME").unwrap();
             let dir_builder = std::fs::DirBuilder::new();
             let mut path = env::temp_dir();
-            path.push(format!("{}-{}", name, epoch.subsec_nanos()));
+            path.push(format!("{name}-{}", epoch.subsec_nanos()));
             dir_builder.create(&path).map(|_| Self { path, cwd })
         }
     }
@@ -866,11 +849,12 @@ mod tests {
     // Test detecting path traversal attacks like CVE-2015-1197
     #[test]
     fn test_read_cpio_and_extract_path_traversal() {
-        let mut file = File::open("tests/path-traversal.cpio").unwrap();
+        let _lock = TEST_LOCK.lock().unwrap();
+        let mut archive = File::open("tests/path-traversal.cpio").unwrap();
         let tempdir = TempDir::new().unwrap();
         set_current_dir(&tempdir.path).unwrap();
-        let got =
-            read_cpio_and_extract(&mut file, &tempdir.path, false, LOG_LEVEL_WARNING).unwrap_err();
+        let got = read_cpio_and_extract(&mut archive, &tempdir.path, false, LOG_LEVEL_WARNING)
+            .unwrap_err();
         assert_eq!(got.kind(), ErrorKind::InvalidData);
         assert_eq!(got.to_string(), format!(
             "The parent directory of \"tmp/trav.txt\" (resolved to \"/tmp\") is not within the directory {:#?}.",
@@ -895,42 +879,43 @@ mod tests {
         assert_eq!(align_to_4_bytes(32), 0);
     }
 
-    #[test]
-    fn test_decompress_program_not_found() {
-        let file = File::open("tests/single.cpio").expect("test cpio should be present");
-        let mut cmd = Command::new("non-existing-program");
-        let got = decompress(&mut cmd, file).unwrap_err();
-        assert_eq!(got.kind(), ErrorKind::Other);
-        assert_eq!(
-            got.to_string(),
-            "Program 'non-existing-program' not found in PATH."
-        );
-    }
-
     #[test]
     fn test_get_cpio_archive_count_single() {
-        let mut file = File::open("tests/single.cpio").expect("test cpio should be present");
-        let count = get_cpio_archive_count(&mut file).unwrap();
+        let _lock = TEST_LOCK.lock().unwrap();
+        let mut archive = File::open("tests/single.cpio").expect("test cpio should be present");
+        let count = get_cpio_archive_count(&mut archive).unwrap();
         assert_eq!(count, 1);
     }
 
+    #[test]
+    fn test_extract_cpio_archive_with_subdir() {
+        let _lock = TEST_LOCK.lock().unwrap();
+        let archive = File::open("tests/single.cpio").unwrap();
+        let tempdir = TempDir::new().unwrap();
+        set_current_dir(&tempdir.path).unwrap();
+        extract_cpio_archive(archive, false, Some("cpio".into()), LOG_LEVEL_WARNING).unwrap();
+        let path = tempdir.path.join("cpio1/path/file");
+        assert!(path.exists());
+    }
+
     #[test]
     fn test_print_cpio_archive_count() {
-        let mut file = File::open("tests/zstd.cpio").expect("test cpio should be present");
+        let _lock = TEST_LOCK.lock().unwrap();
+        let mut archive = File::open("tests/zstd.cpio").expect("test cpio should be present");
         let mut output = Vec::new();
 
-        let count = get_cpio_archive_count(&mut file).unwrap();
+        let count = get_cpio_archive_count(&mut archive).unwrap();
         assert_eq!(count, 2);
 
-        file.seek(SeekFrom::Start(0)).unwrap();
-        print_cpio_archive_count(file, &mut output).unwrap();
+        archive.seek(SeekFrom::Start(0)).unwrap();
+        print_cpio_archive_count(archive, &mut output).unwrap();
         assert_eq!(String::from_utf8(output).unwrap(), "2\n");
     }
 
     #[test]
     fn test_read_cpio_and_print_long_format_character_device() {
         // Wrapped before mtime and filename
-        let cpio_data = b"07070100000003000021A4000000000000\
+        let archive = b"07070100000003000021A4000000000000\
         00000000000167055BC800000000000000000000000000000005000000010000\
         000C00000000dev/console\0\0\0\
         0707010000000000000000000000000000000000000001\
@@ -941,10 +926,11 @@ mod tests {
         env::set_var("TZ", "UTC");
         unsafe { tzset() };
         read_cpio_and_print_long_format(
-            &mut cpio_data.as_ref(),
+            &mut archive.as_ref(),
             &mut output,
             1728486311,
             &mut user_group_cache,
+            false,
         )
         .unwrap();
         assert_eq!(
@@ -956,7 +942,7 @@ mod tests {
     #[test]
     fn test_read_cpio_and_print_long_format_directory() {
         // Wrapped before mtime and filename
-        let cpio_data = b"07070100000001000047FF000000000000007B00000002\
+        let archive = b"07070100000001000047FF000000000000007B00000002\
         66A6E40400000000000000000000000000000000000000000000000B00000000\
         /var/crash\0\0\0\0\
         0707010000000000000000000000000000000000000001\
@@ -968,10 +954,11 @@ mod tests {
         env::set_var("TZ", "UTC");
         unsafe { tzset() };
         read_cpio_and_print_long_format(
-            &mut cpio_data.as_ref(),
+            &mut archive.as_ref(),
             &mut output,
             1722389471,
             &mut user_group_cache,
+            false,
         )
         .unwrap();
         assert_eq!(
@@ -983,7 +970,7 @@ mod tests {
     #[test]
     fn test_read_cpio_and_print_long_format_file() {
         // Wrapped before mtime and filename
-        let cpio_data = b"070701000036E4000081A4000003E8000007D000000001\
+        let archive = b"070701000036E4000081A4000003E8000007D000000001\
         66A3285300000041000000000000002400000000000000000000000D00000000\
         conf/modules\0\0\
         linear\nmultipath\nraid0\nraid1\nraid456\nraid5\nraid6\nraid10\nefivarfs\0\0\0\0\
@@ -996,10 +983,11 @@ mod tests {
         env::set_var("TZ", "UTC");
         unsafe { tzset() };
         read_cpio_and_print_long_format(
-            &mut cpio_data.as_ref(),
+            &mut archive.as_ref(),
             &mut output,
             1722645915,
             &mut user_group_cache,
+            false,
         )
         .unwrap();
         assert_eq!(
@@ -1011,7 +999,7 @@ mod tests {
     #[test]
     fn test_read_cpio_and_print_long_format_symlink() {
         // Wrapped before mtime and filename
-        let cpio_data = b"0707010000000D0000A1FF000000000000000000000001\
+        let archive = b"0707010000000D0000A1FF000000000000000000000001\
         6237389400000007000000000000000000000000000000000000000400000000\
         bin\0\0\0usr/bin\0\
         0707010000000000000000000000000000000000000001\
@@ -1021,10 +1009,11 @@ mod tests {
         let mut user_group_cache = UserGroupCache::new();
         user_group_cache.insert_test_data();
         read_cpio_and_print_long_format(
-            &mut cpio_data.as_ref(),
+            &mut archive.as_ref(),
             &mut output,
             1722645915,
             &mut user_group_cache,
+            false,
         )
         .unwrap();
         assert_eq!(
@@ -1033,13 +1022,45 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_read_cpio_and_print_long_format_print_ino() {
+        // Wrapped after mtime
+        let archive = b"07070100000000000041ED00000000000000000000000265307180\
+        00000000000000000000000000000000000000000000000200000000.\0\
+        07070100000001000041ED00000000000000000000000265307180\
+        00000000000000000000000000000000000000000000000700000000kernel\0\0\0\0\
+        070701000000000000000000000000000000000000000100000000\
+        00000000000000000000000000000000000000000000000B00000000TRAILER!!!\0\0\0\0";
+        let mut output = Vec::new();
+        let mut user_group_cache = UserGroupCache::new();
+        user_group_cache.insert_test_data();
+        env::set_var("TZ", "UTC");
+        unsafe { tzset() };
+        read_cpio_and_print_long_format(
+            &mut archive.as_ref(),
+            &mut output,
+            1722645915,
+            &mut user_group_cache,
+            true,
+        )
+        .unwrap();
+        assert_eq!(
+            String::from_utf8(output).unwrap(),
+            concat!(
+                "   0 drwxr-xr-x   2 root     root            0 Oct 19  2023 .\n",
+                "   1 drwxr-xr-x   2 root     root            0 Oct 19  2023 kernel\n"
+            )
+        );
+    }
+
     #[test]
     fn test_write_character_device() {
+        let _lock = TEST_LOCK.lock().unwrap();
         if getuid() != 0 {
             // This test needs to run as root.
             return;
         }
-        let mut header = Header::new(1, 0o20_644, 0, 0, 0, 1740402179, 0, "./null");
+        let mut header = Header::new(1, 0o20_644, 0, 0, 0, 1740402179, 0, 0, 0, "./null");
         header.rmajor = 1;
         header.rminor = 3;
         write_character_device(&header, true, LOG_LEVEL_WARNING).unwrap();
@@ -1058,6 +1079,7 @@ mod tests {
 
     #[test]
     fn test_write_directory_with_setuid() {
+        let _lock = TEST_LOCK.lock().unwrap();
         let mut mtimes = BTreeMap::new();
         let header = Header::new(
             1,
@@ -1067,6 +1089,8 @@ mod tests {
             0,
             1720081471,
             0,
+            0,
+            0,
             "./directory_with_setuid",
         );
         write_directory(&header, true, LOG_LEVEL_WARNING, &mut mtimes).unwrap();
@@ -1085,6 +1109,7 @@ mod tests {
 
     #[test]
     fn test_write_file_with_setuid() {
+        let _lock = TEST_LOCK.lock().unwrap();
         let mut seen_files = SeenFiles::new();
         let header = Header::new(
             1,
@@ -1094,6 +1119,8 @@ mod tests {
             0,
             1720081471,
             9,
+            0,
+            0,
             "./file_with_setuid",
         );
         let cpio = b"!/bin/sh\n\0\0\0";
@@ -1118,6 +1145,7 @@ mod tests {
 
     #[test]
     fn test_write_symbolic_link() {
+        let _lock = TEST_LOCK.lock().unwrap();
         let header = Header::new(
             1,
             0o120_777,
@@ -1126,6 +1154,8 @@ mod tests {
             0,
             1721427072,
             12,
+            0,
+            0,
             "./dead_symlink",
         );
         let cpio = b"/nonexistent";
