@@ -10,20 +10,21 @@ use std::process::ExitCode;
 use lexopt::prelude::*;
 
 use threecpio::{
-    examine_cpio_content, extract_cpio_archive, list_cpio_content, print_cpio_archive_count,
-    LOG_LEVEL_DEBUG, LOG_LEVEL_INFO, LOG_LEVEL_WARNING,
+    create_cpio_archive, examine_cpio_content, extract_cpio_archive, list_cpio_content,
+    print_cpio_archive_count, LOG_LEVEL_DEBUG, LOG_LEVEL_INFO, LOG_LEVEL_WARNING,
 };
 
 #[derive(Debug)]
 struct Args {
     count: bool,
+    create: bool,
     directory: String,
     examine: bool,
     extract: bool,
     force: bool,
     list: bool,
     log_level: u32,
-    file: String,
+    archive: Option<String>,
     preserve_permissions: bool,
     subdir: Option<String>,
 }
@@ -32,13 +33,15 @@ fn print_help() {
     let executable = std::env::args().next().unwrap();
     println!(
         "Usage:
-    {executable} --count FILE
-    {executable} {{-e|--examine}} FILE
-    {executable} {{-t|--list}} [-v] FILE
-    {executable} {{-x|--extract}} [-v|--debug] [-C DIR] [-p] [-s NAME] [--force] FILE
+    {executable} --count ARCHIVE
+    {executable} {{-c|--create}} [-v|--debug] [-C DIR] [ARCHIVE] < manifest
+    {executable} {{-e|--examine}} ARCHIVE
+    {executable} {{-t|--list}} [-v|--debug] ARCHIVE
+    {executable} {{-x|--extract}} [-v|--debug] [-C DIR] [-p] [-s NAME] [--force] ARCHIVE
 
 Optional arguments:
   --count        Print the number of concatenated cpio archives.
+  -c, --create   Create a new cpio archive from the manifest on stdin.
   -e, --examine  List the offsets of the cpio archives and their compression.
   -t, --list     List the contents of the cpio archives.
   -x, --extract  Extract cpio archives.
@@ -59,11 +62,12 @@ Optional arguments:
 fn print_version() {
     let name = std::option_env!("CARGO_BIN_NAME").unwrap();
     let version = std::option_env!("CARGO_PKG_VERSION").unwrap();
-    println!("{} {}", name, version);
+    println!("{name} {version}");
 }
 
 fn parse_args() -> Result<Args, lexopt::Error> {
     let mut count = 0;
+    let mut create = 0;
     let mut examine = 0;
     let mut extract = 0;
     let mut force = false;
@@ -71,7 +75,7 @@ fn parse_args() -> Result<Args, lexopt::Error> {
     let mut list = 0;
     let mut log_level = LOG_LEVEL_WARNING;
     let mut directory = ".".into();
-    let mut file = None;
+    let mut archive = None;
     let mut subdir: Option<String> = None;
     let mut parser = lexopt::Parser::from_env();
     while let Some(arg) = parser.next()? {
@@ -79,6 +83,9 @@ fn parse_args() -> Result<Args, lexopt::Error> {
             Long("count") => {
                 count = 1;
             }
+            Short('c') | Long("create") => {
+                create = 1;
+            }
             Short('C') | Long("directory") => {
                 directory = parser.value()?.string()?;
             }
@@ -116,32 +123,39 @@ fn parse_args() -> Result<Args, lexopt::Error> {
             Short('x') | Long("extract") => {
                 extract = 1;
             }
-            Value(val) if file.is_none() => {
-                file = Some(val.string()?);
+            Value(val) if archive.is_none() => {
+                archive = Some(val.string()?);
             }
             _ => return Err(arg.unexpected()),
         }
     }
 
-    if count + examine + extract + list != 1 {
-        return Err("Either --count, --examine, --extract, or --list must be specified!".into());
+    if count + create + examine + extract + list != 1 {
+        return Err(
+            "Either --count, --create, --examine, --extract, or --list must be specified!".into(),
+        );
     }
 
     if let Some(ref s) = subdir {
         if s.contains('/') {
-            return Err(format!("Subdir '{}' must not contain slashes!", s).into());
+            return Err(format!("Subdir '{s}' must not contain slashes!").into());
         }
     }
 
+    if create != 1 && archive.is_none() {
+        return Err("missing argument ARCHIVE".into());
+    }
+
     Ok(Args {
         count: count == 1,
+        create: create == 1,
         directory,
         examine: examine == 1,
         extract: extract == 1,
         force,
         list: list == 1,
         log_level,
-        file: file.ok_or("missing argument FILE")?,
+        archive,
         preserve_permissions,
         subdir,
     })
@@ -159,27 +173,26 @@ fn is_root() -> bool {
 fn create_and_set_current_dir(path: &str, force: bool) -> Result<(), String> {
     if let Err(e) = set_current_dir(path) {
         if e.kind() != ErrorKind::NotFound {
-            return Err(format!("Failed to change directory to '{}': {}", path, e));
+            return Err(format!("Failed to change directory to '{path}': {e}"));
         }
         if let Err(e) = create_dir(path) {
-            return Err(format!("Failed to create directory '{}': {}", path, e));
+            return Err(format!("Failed to create directory '{path}': {e}"));
         }
         if let Err(e) = set_current_dir(path) {
-            return Err(format!("Failed to change directory to '{}': {}", path, e));
+            return Err(format!("Failed to change directory to '{path}': {e}"));
         }
     }
     if !force {
         match is_empty_directory(".") {
             Err(e) => {
                 return Err(format!(
-                    "Failed to check content of directory '{}': {}",
-                    path, e
+                    "Failed to check content of directory '{path}': {e}"
                 ));
             }
             Ok(false) => {
                 return Err(format!(
-                    "Target directory '{}' is not empty. Use --force to overwrite existing files!",
-                    path
+                    "Target directory '{path}' is not empty. \
+                    Use --force to overwrite existing files!",
                 ));
             }
             Ok(true) => {}
@@ -193,17 +206,54 @@ fn main() -> ExitCode {
     let args = match parse_args() {
         Ok(a) => a,
         Err(e) => {
-            eprintln!("{}: Error: {}", executable, e);
+            eprintln!("{executable}: Error: {e}");
             return ExitCode::from(2);
         }
     };
 
-    let file = match File::open(&args.file) {
+    if args.create {
+        let mut archive = None;
+        if let Some(path) = args.archive.as_ref() {
+            archive = match File::create(path) {
+                Ok(f) => Some(f),
+                Err(e) => {
+                    eprintln!("{executable}: Error: Failed to create '{path}': {e}");
+                    return ExitCode::FAILURE;
+                }
+            };
+            if args.log_level >= LOG_LEVEL_DEBUG {
+                eprintln!("{executable}: Opened '{path}' for writing.");
+            }
+        }
+        if let Err(e) = set_current_dir(&args.directory) {
+            eprintln!(
+                "{executable}: Error: Failed to change directory to '{}': {e}",
+                args.directory,
+            );
+            return ExitCode::FAILURE;
+        }
+        let result = create_cpio_archive(archive, args.log_level);
+        if let Err(error) = result {
+            match error.kind() {
+                ErrorKind::BrokenPipe => {}
+                _ => {
+                    eprintln!(
+                        "{executable}: Error: Failed to create '{}': {error}",
+                        args.archive.unwrap_or("cpio on stdout".into()),
+                    );
+                    return ExitCode::FAILURE;
+                }
+            }
+        }
+        return ExitCode::SUCCESS;
+    };
+
+    let archive = match File::open(args.archive.as_ref().unwrap()) {
         Ok(f) => f,
         Err(e) => {
             eprintln!(
-                "{}: Error: Failed to open '{}': {}",
-                executable, args.file, e
+                "{executable}: Error: Failed to open '{}': {e}",
+                args.archive.unwrap(),
             );
             return ExitCode::FAILURE;
         }
@@ -211,7 +261,7 @@ fn main() -> ExitCode {
 
     if args.extract {
         if let Err(e) = create_and_set_current_dir(&args.directory, args.force) {
-            eprintln!("{}: Error: {}", executable, e);
+            eprintln!("{executable}: Error: {e}");
             return ExitCode::FAILURE;
         }
     }
@@ -220,19 +270,27 @@ fn main() -> ExitCode {
     let (operation, result) = if args.count {
         (
             "count number of cpio archives",
-            print_cpio_archive_count(file, &mut stdout),
+            print_cpio_archive_count(archive, &mut stdout),
         )
     } else if args.examine {
-        ("examine content", examine_cpio_content(file, &mut stdout))
+        (
+            "examine content",
+            examine_cpio_content(archive, &mut stdout),
+        )
     } else if args.extract {
         (
             "extract content",
-            extract_cpio_archive(file, args.preserve_permissions, args.subdir, args.log_level),
+            extract_cpio_archive(
+                archive,
+                args.preserve_permissions,
+                args.subdir,
+                args.log_level,
+            ),
         )
     } else if args.list {
         (
             "list content",
-            list_cpio_content(file, &mut stdout, args.log_level),
+            list_cpio_content(archive, &mut stdout, args.log_level),
         )
     } else {
         unreachable!("no operation specified");
@@ -243,8 +301,8 @@ fn main() -> ExitCode {
             ErrorKind::BrokenPipe => {}
             _ => {
                 eprintln!(
-                    "{}: Error: Failed to {} of '{}': {}",
-                    executable, operation, args.file, e
+                    "{executable}: Error: Failed to {operation} of '{}': {e}",
+                    args.archive.unwrap(),
                 );
                 return ExitCode::FAILURE;
             }
